#!/usr/bin/python3

import rospy
import numpy as np
from sensor_msgs.msg import NavSatFix
from sensor_msgs.msg import Imu
from geometry_msgs.msg import PointStamped
from geometry_msgs.msg import QuaternionStamped
from dji_osdk_ros.msg import VOPosition
from geometry_msgs.msg import Transform
from geometry_msgs.msg import Vector3
from geometry_msgs.msg import Quaternion
from scipy.spatial.transform import Rotation
from tools import lat_lon2meters, meters2lat_lon
import math

homeSetFlag = False
homePoint = NavSatFix()
transform = Transform()
quat = Quaternion()
translation = Vector3()
posePublisher = rospy.Publisher("pose_feedback", Transform, queue_size = 1)
angleOffset = 0.0

def point_rotation(origin, pt, ang):
    # returns the pt rotated about the origin by ang (ang in degrees)
    c = math.cos(math.radians(ang))
    s = math.sin(math.radians(ang))
    # translate to origin
    pt_temp = [pt[0] - origin[0], pt[1] - origin[1]]
    pt_spun = [ pt_temp[0]*c - pt_temp[1]*s, pt_temp[0]*s + pt_temp[1]*c ]
    # translate back to frame
    pt_spun = [pt_spun[0] + origin[0], pt_spun[1] + origin[1]]
    return pt_spun


def rotCallback(data):
    global quat
    quat = data.quaternion;

def posCallback(data):
    global quat
    # convert and compare against home point
    homePoint = rospy.get_param("~anchorPoint")
    homePointMeters = lat_lon2meters([homePoint[0], homePoint[1]])
    currentPointMeters = lat_lon2meters([data.latitude, data.longitude])
    currentOffset = [currentPointMeters[0] - homePointMeters[0], currentPointMeters[1] - homePointMeters[1]]
    angleOffset = rospy.get_param("~angleOffset")
    currentRotatedOffset = point_rotation([0,0], currentOffset, angleOffset)

    # performing dark magics
    translation = Vector3()
    translation.x = currentRotatedOffset[1]
    translation.y = currentRotatedOffset[0]
    translation.z = data.altitude - homePoint[2]

    transform.translation = translation
    transform.rotation = quat
    posePublisher.publish(transform)

def main():
    rospy.init_node("poseFeedbackConversion")
    rospy.Subscriber("dji_osdk_ros/gps_position", NavSatFix, posCallback, queue_size = 1)
    rospy.Subscriber("dji_osdk_ros/attitude", QuaternionStamped, rotCallback, queue_size = 1)
    rospy.spin()
    # get anchor point from params

if __name__ == '__main__':
    main()
