# NVIDIA Jetson Xavier NX Setup

Below are the steps required to set up a stock Jetson to use this code base:

* Set up the SD card to have an OS on it, either by:
    - Using NVIDIA's set up procedure found [here](https://developer.nvidia.com/embedded/learn/get-started-jetson-xavier-nx-devkit#prepare), which then would require a lot of installations/additional set up (OpenCV, Ros, etc.).
    - Using the CAS Lab's own images found on the [drive](https://drive.google.com/file/d/16X103JHvNeeOU7jYVqtzIdrLQNcEMMz-/view?usp=sharing).

* If using the NVIDIA option a few more things need to be configured:
    - Follow the steps listed in the DJI Onboard SDK software set up page [here](https://developer.dji.com/onboard-sdk/documentation/development-workflow/environment-setup.html) (also can be helpful for debugging, a complete list of vehicle activation checks can be found [here](https://forum.dji.com/thread-216529-1-1.html))
    - The udev rules file is slightly different than what DJI suggests, this is because the USB device name needs to be static so it can be found every time. The udev file contents below is designed for use with [this](https://www.adafruit.com/product/954) USB device. To use a different device the `idVendor` field can be found by inspecting `dmesg` or `lsusb`

* The hosts and hostname will need to be configured regardless of the image source
    - The `/etc/hosts` template, listed below, makes it easier to ssh into the Jetson and launch ROS nodes.
    - The `/etc/hostname` file holds the name of the device, to be used in the hosts file and for ssh-ing


### Cloning The CAS Lab Jetson Image

To set up a new SD card using the image on the drive, follow the steps below. Instead of using a loopback device you can just use another SD from an already set up Jetson, in which case the loopback stuff can be skipped. Disclaimer, these commands are pretty dangerous if mistyped and the original contents of the SD will be deleted of course. Also this process needs to be done on a linux machine with a lot of free space. The majority of these steps were taken from this one random forum [post](https://forums.developer.nvidia.com/t/shink-system-image-of-jetson-nano-for-smaller-size-sd-card/80522/16).

* Prepare the image on a loopback device
  - Find the next free loopback device with `losetup -f`, this gives you something like `/dev/loop12`. We'll keep using `/dev/loop12` but yours may differ.
  - Link the loopback device with the image in question by `losetup /dev/loop12 path/to/neruda_backup.img`, the second argument being the CAS Lab Jetson image.
  - Check out the partition table of the image `fdisk -l`, there should be 12 different partitions with the first taking the majority of space on the image. 12 here being the same as the loopback device is just a coincidence.
  - After all the other steps are completed, remove the loopback device with `losetup -d /dev/loop12`

* Copy partition table to SD
  - Open the loopback device (or SD card, if cloning an already set up card) with `gdisk /dev/loop12`
  - Enter expert command mode with `x` (can use `?` to get commands)
  - Select `copy partition table to new device` and type `/dev/loop12` in the prompt, this may throw warnings but they are likely to do with overwriting existing data
  - Write changes to disk with `w` and `y` in the prompt, exit gdisk if no errors.
  - Now we can copy the contents of each partition to the target drive, which we'll call `/dev/sdb` but yours may differ. Find target SD card with `fdisk -l` with and without the card plugged in. Run `dd if=/dev/loop12p1 of=/dev/sdb1` for each partition in `/dev/loop12` of which there are 12 (tedious as heck, sorry). The target SD card now should be a perfect clone of the source.

### UDEV Rules File

The udev rules contents, to be placed in `/etc/udev/rules.d/DJIDevice.rules`, is listed below

```
SUBSYSTEM=="usb", ATTRS{idVendor}=="2ca3", MODE="0666"
SUBSYSTEM=="tty", ATTRS{idVendor}=="10c4", ATTRS{idProduct}=="ea60", SYMLINK+="dji_usb"
```

### Hosts File

The hosts contents, to be placed in `/etc/hosts` is listed below

```
127.0.0.1       localhost
192.168.1.100   pop-os
192.168.1.230   pop-os
127.0.0.1       yeats
192.168.1.101   neruda
192.168.1.191   nerudaHouse
192.168.1.102   plath
192.168.1.24    plathHouse
192.168.1.103   yeats # --> your hostname here, house denotes local home network
192.168.1.198   yeatsHouse # --> these house IPs are from a random home router, change if using

# The following lines are desirable for IPv6 capable hosts
::1     ip6-localhost ip6-loopback
fe00::0 ip6-localnet
ff00::0 ip6-mcastprefix
ff02::1 ip6-allnodes
ff02::2 ip6-allrouters
```
---

# Hardware Setup

Procedure for drone-Jetson communication is found [here](https://developer.dji.com/onboard-sdk/documentation/development-workflow/hardware-setup.html) note details on Advanced Sensing as they are required to use drone cameras.

# Running via Launch File

To test the packages, a DJI M200 series drone is required. The hardware setup must be correct or errors will occur due to the Jetson not being able to communicate with the drone. The DJI Assistant simulator (available from the DJI [website](https://www.dji.com/downloads/softwares/assistant-dji-2-for-matrice)) is helpful for desktop testing. The simulator requires a USB connection to the drone and the simulator's `API Control` option needs to be checked. 

Once the drone is powered on and an ssh connection is made, the nodes can be launched via

`roslaunch dji_osdk_ros local_test.launch`

Which launches the vehicle node, various position monitoring nodes, and the main control node `bus_driver`. Drone actions can be carried out via `rosservice call /overwatch <action>` where `<action>` is an integer. The various actions are listed in `rossrv show dji_osdk_ros/overwatch`. 

# Branches

There are multiple branches in this repository, each designed for use with different versions of the M210 drone. The main branch, `replicants`, is made to work with the DJI M210V2 drone.

# Debugging

Why it doesn't work??

* Check that the SD card is pushed all the way into the slot, this happens quite a lot
* Try the SD card in a different Jetson to see if the Jetson is to blame
* Check that the power LED on the board is turning green when plugged in
* Wifi is likely not set up correctly after cloning (MAC address conflicts happen), plugging in an HDMI and keyboard can be helpful, rather than going headless

### Vehicle Activation Errors

Check [here](https://forum.dji.com/thread-216529-1-1.html).
